/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010.
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *
 *  Authors:
 *  2007-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl>
 *
 */



/*!
    \page lcmaps_scas_client.mod Site Central Authorization Service client plugin

    \section posixsynopsis SYNOPSIS
    \b lcmaps_scas_client.mod

    \section posixdesc DESCRIPTION

description (need to fill in)


*/


/*!
    \file   lcmaps_scas_client.c
    \brief  Plugin to contact the Site Central Authorization Service
    \author Oscar Koeroo for the EGEE.
*/



/* Define both _XOPEN_SOURCE for strdup and _GNU_SOURCE, this makes Solaris
 * happier */
#define _XOPEN_SOURCE   600
/* Try to provide RTLD_DEFAULT */
/* MacOS needs no POSIX or _DARWIN_C_SOURCE to have RTLD_DEFAULT */
#ifdef __APPLE__
# define _DARWIN_C_SOURCE
#else
# define _GNU_SOURCE
#endif
#include <dlfcn.h>

/*****************************************************************************
                            Include header files
******************************************************************************/
#include "lcmaps_scas_config.h"
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <openssl/x509.h>
#include <openssl/x509v3.h>
#include <openssl/x509_vfy.h>
#include <openssl/asn1.h>
#include <time.h>
#include <fcntl.h>
#include <errno.h>
#include <limits.h>
#include <strings.h>

#include <sys/stat.h>
#include <unistd.h>

#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>


#include <lcmaps/lcmaps_modules.h>
#include <lcmaps/lcmaps_arguments.h>
#include <lcmaps/lcmaps_cred_data.h>
/*    #include <lcmaps/lcmaps_voms_utils.h> */
#include <lcmaps/lcmaps_vo_data.h>
#include <lcmaps/lcmaps_types.h>

#if defined(HAVE_LCMAPS_LCMAPS_PLUGIN_PROTOTYPE_H)
#   include <lcmaps/lcmaps_plugin_prototypes.h>
#else
#   include "lcmaps_plugin_prototypes.h"
#endif
/*    #include <voms/voms_apic.h> */

/* lcmaps_basic.h for lcmaps_is_set_to_verify_voms_attributes prototype */
#include <lcmaps/lcmaps_basic.h>

/* #undef TESTBIO */
/* #define TESTBIO */
#ifdef TESTBIO
#    include <openssl/pem.h>
#    include <openssl/bio.h>
#endif

#include "parameters.h"


/*** SAML2-XACML2 ***/
#include "xacml_client.h"
#include "xacml_io_ssl.h"
#include "net_common.h"
#include "ssl-common.h"

#include "pep_obligation_handlers.h"
#include "xacml_attr_obl_profile.h"

/*** SAML2-XACML2 ***/


/******************************************************************************
                                Definitions
******************************************************************************/

#define PLUGIN_RUN      0
#define PLUGIN_VERIFY   1


/* #define VOMS_BUFFER_SIZE  1024 */

/* The next buffer is used for logging the XACML output */
#define PIPE_BUFFER_SIZE ((size_t)1024)

/* What do we want to do with this time presentation? */
/*#define GMT               1*/

/*#define LEAVE_TIME_ITS_OK 3*/

/* Define default CA certificate directory */
#ifndef DEFAULT_CERT_DIR
# define DEFAULT_CERT_DIR   "/etc/grid-security/certificates/"
#endif

#define TIME_BUF_LEN	21

typedef enum ENDPOINT_SELECTION_E {
    ENDPOINT_SELECTION_ROUND_ROBIN_TOP_TO_BOTTOM = 1,
    ENDPOINT_SELECTION_ROUND_ROBIN_RANDOM_START  = 2,
    ENDPOINT_SELECTION_RANDOM                    = 3
} ENDPOINT_SELECTION_T;


typedef enum XACML_QUERY_ANSWER_E {
    XACML_QUERY_ALLOWED          = 0,
    XACML_QUERY_BANNED           = 1,
    XACML_QUERY_PROTOCOL_ERROR   = 2
} XACML_QUERY_ANSWER_T;

/******************************************************************************
                          Module specific prototypes
******************************************************************************/

static int plugin_run_or_verify(int, lcmaps_argument_t *, int);

static int pem_string_to_x509(X509 **px, char *certstring);
static XACML_QUERY_ANSWER_T xacmlqueryscas (const char *           ,
                                            xacml_request_t  ,
                                            xacml_response_t ,
                                            FILE *,
					    int ,
					    int *,
					    const char * ,
					    const char * ,
					    unsigned int );
static int parse_and_store_creddata(const char *endpoint);
static void log_fake_stderr(int level, int fd);

static int lcmaps_scas_client_time_t_to_string(time_t the_time,
					       char *buf,
					       size_t buf_len);

static int fill_subject_attributes(subject_attr_t *d, STACK_OF(X509) *myChain);
static void free_subject_attributes(subject_attr_t *d);

static char *get_serialNumber(X509 *cert);
static char *stack_of_x509_to_pem(STACK_OF(X509) *myChain);

/******************************************************************************
                       Define module specific variables
******************************************************************************/

#define MAX_ENDPOINTS         32
#define MAX_ENDPOINT_BUFSIZE  ((size_t)256)

/*
static int    endpoint_selection_strategy = ENDPOINT_SELECTION_ROUND_ROBIN_TOP_TO_BOTTOM;
static int    endpoint_selection_strategy = ENDPOINT_SELECTION_ROUND_ROBIN_RANDOM_START;
*/
static ENDPOINT_SELECTION_T endpoint_selection_strategy = ENDPOINT_SELECTION_ROUND_ROBIN_RANDOM_START;
static char   endpoints[MAX_ENDPOINTS][MAX_ENDPOINT_BUFSIZE];
static unsigned int n_endpoints = 0;
static unsigned int endpoint_retrycount = 2; /* retry each endpoint twice */
static int    endpoint_retrycount_is_set = 0;
static const char * resourcetype = NULL;
static const char * actiontype = NULL;
/*
char * certFilePub  = NULL;
char * certFilePriv = NULL;
char * ca_path      = NULL;
*/

static int setResourceX509ID = 0;
static int use_pool_index_bug_fix = 0;

static int use_x509_chain = 1;  /* default: use px509_chain as input for user
				   credentials, alternative is to use the DN and
				   FQANs directly, but this results in less
				   information being sent to the SCAS (e.g.
				   notBefore/notAfter and issuer_dn are not
				   being sent) */

static int cert_client_side_verified = -1; /* default: unknown whether proxy
					      is verified on client side */

static int use_cert_chain_attribute = 0; /* default: don't send the cert-chain
					    subject attribute to the server */

static int use_dn_from_credential_data = 0; /* default: don't use registered
					       credential data for subjectDN
					       but take from input PEM string */

/*****************************************************************************/




/******************************************************************************
Function:   plugin_initialize
Description:
    Initialize plugin
Parameters:
    argc, argv
    argv[0]: the name of the plugin
Returns:
    LCMAPS_MOD_SUCCESS : succes
    LCMAPS_MOD_FAIL    : failure
    LCMAPS_MOD_NOFILE  : db file not found (will halt LCMAPS initialization)
******************************************************************************/
int plugin_initialize(
        int argc,
        char ** argv
)
{
    const char * logstr = "lcmaps_plugin_scas_client-plugin_initialize()";
    int i;
    unsigned int u;
    struct stat s;

    long long_val;

    /* The new default is to enable the TCP/IP keep-alive feature explicitly */
    xacml_io_set_keepalive(XACML_KEEPALIVE_ENABLED);

    /* Setting defaults either USA or Europe, see configure */
    xacml_io_set_socket_timeout(SOCKET_TIMEOUT);     /* mseconds */
    xacml_io_set_connection_timeout(CONNECTION_TIMEOUT);   /* seconds */
    xacml_io_set_incremental_backoff(INCREMENTAL_BACKOFF); /* mseconds */

    xacml_io_certFilePub  = NULL;
    xacml_io_certFilePriv = NULL;
    xacml_io_certCAPath   = NULL;


    /* connStr = (char *) malloc (sizeof(char) * 10000); */
    /* for (i = 0; i < 1000; i++) connStr[i] = '\0'; */

    lcmaps_log(LOG_DEBUG,"%s: passed arguments:\n", logstr);
    for (i=0; i < argc; i++)
    {
       lcmaps_log(LOG_DEBUG,"%s: arg %d is %s\n", logstr, i, argv[i]);
    }

    /*
     * CERTDIR = The directory which contains the CA certificates
     * VOMSDIR = The directory which contains the certificates of the VOMS servers
     */

    /*
     * Parse arguments, argv[0] = name of plugin, so start with i = 1
     */

    for (i = 1; i < argc; i++)
    {
        /* Endpoint selection strategy */
        if ( ((strcasecmp(argv[i], "--endpoint-strategy") == 0) ||
              (strcasecmp(argv[i], "-endpoint-strategy") == 0)) &&
             (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                if (strcasecmp (argv[i + 1], "round-robin") == 0)
                    endpoint_selection_strategy = ENDPOINT_SELECTION_ROUND_ROBIN_TOP_TO_BOTTOM;
                else if (strcasecmp (argv[i + 1], "round-robin-random-start") == 0)
                    endpoint_selection_strategy = ENDPOINT_SELECTION_ROUND_ROBIN_RANDOM_START;
                else if (strcasecmp (argv[i + 1], "random") == 0)
                    endpoint_selection_strategy = ENDPOINT_SELECTION_RANDOM;
                else
                {
                    lcmaps_log(LOG_ERR, "%s: Error: --endpoint-strategy not understood. Options are: \"round-robin\", \"round-robin-random-start\" and \"random\".\n", logstr);
                    return LCMAPS_MOD_FAIL;
                }
            }
            i++;
        }

        /* Endpoint */
        else if ( ( (strcasecmp(argv[i], "-endpoint") == 0) ||
                    (strcasecmp(argv[i], "--endpoint") == 0) ) &&
                  (i + 1 < argc) )
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                if (n_endpoints < MAX_ENDPOINTS)
                {
/*                    bzero (endpoints[n_endpoints], MAX_ENDPOINT_BUFSIZE);*/
		    memset(endpoints[n_endpoints],'\0',MAX_ENDPOINT_BUFSIZE);
                    if (strlen (argv[i + 1]) >= MAX_ENDPOINT_BUFSIZE)
                    {
                        lcmaps_log(LOG_ERR, "%s: Error: endpoint URL is too long: %s...\n", strncpy(endpoints[n_endpoints], argv[i + 1], MAX_ENDPOINT_BUFSIZE - 1), logstr);
                        return LCMAPS_MOD_FAIL;
                    }
                    else
                    {
                        strncpy(endpoints[n_endpoints], argv[i + 1], MAX_ENDPOINT_BUFSIZE - 1);
                        n_endpoints++;
                    }
                }
                else
                {
                    lcmaps_log(LOG_ERR, "%s: Error: Too many endpoints configured, the maximum amount of configurable endpoints is %d\n", logstr, MAX_ENDPOINTS);
                    return LCMAPS_MOD_FAIL;
                }
            }
            i++;
        }

        /* Change the privileges of the process to this user account to read a proxy or certificate files from the file system */
        else if ( ((strcasecmp(argv[i], "-cert-owner")  == 0) ||
                   (strcasecmp(argv[i], "--cert-owner") == 0)) && (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                xacml_io_certOwner = strdup(argv[i + 1]);
            }
            i++;
        }

        /* Endpoint retry count */
        else if ( ( (strcasecmp(argv[i], "-retry") == 0) ||
                    (strcasecmp(argv[i], "--retry") == 0) ) &&
                  (i + 1 < argc) )
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
		errno=0;
		long_val = strtol (argv[i + 1], NULL, 10);
                if (errno == EINVAL)
                {
                    lcmaps_log(LOG_ERR, "%s: Error: Couldn't parse the value \"%s\" to a number for the --retry option.\n", logstr, argv[i + 1]);
                    return LCMAPS_MOD_FAIL;
                }
		if ((long unsigned)long_val>UINT_MAX)
		{
		    lcmaps_log(LOG_ERR, "%s: Error: Endpoint retry count (%ld) is larger than UINT_MAX.\n", logstr, long_val);
		    return LCMAPS_MOD_FAIL;
		}
		if (long_val<=0)
		{
                    lcmaps_log(LOG_ERR, "%s: Error: Endpoint retry count to low. It must be at least '1'\n", logstr);
                    return LCMAPS_MOD_FAIL;
                }
		endpoint_retrycount=(unsigned int)long_val;
		endpoint_retrycount_is_set=1;
            }
            i++;
        }


        /* (optional) Host certificate file */
        else if ( ( (strcasecmp(argv[i], "-cert") == 0) ||
		    (strcasecmp(argv[i], "--cert") == 0))
                 && (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                xacml_io_certFilePub = strdup(argv[i + 1]);
            }
            i++;
        }

        /* (optional) Host certificate private key file */
        else if ( ( (strcasecmp(argv[i], "-key") == 0) ||
                    (strcasecmp(argv[i], "--key") == 0)) &&
		  (i + 1 < argc) )
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                xacml_io_certFilePriv = strdup(argv[i + 1]);
            }
            i++;
        }

        /* CA Path - path to CA certificate files (and CRLs) */
        else if ( ( (strcasecmp(argv[i], "-capath") == 0) ||
                    (strcasecmp(argv[i], "--capath") == 0)) &&
		  (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                /* Check if I can stat() the CA directory */
                if (stat (argv[i + 1], &s) < 0)
                {
                    lcmaps_log(LOG_ERR,"%s: Error: The CA certificates path \"%s\" doesn't exist or is at least not stat()-able\n", logstr, argv[i + 1]);
                    return LCMAPS_MOD_FAIL;
                }

                xacml_io_certCAPath = strdup(argv[i + 1]);
            }
            i++;
        }

        /* Resource Type - Set the resource type to CE, WN, RB, SE */
        else if ( ( (strcasecmp(argv[i], "-resourcetype") == 0) ||
                    (strcasecmp(argv[i], "--resourcetype") == 0)) &&
		  (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                char * tmprt = argv[i + 1];

                if (strcasecmp (tmprt, "se") == 0)      resourcetype = XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_TYPE_SE;
                else if (strcasecmp (tmprt, "ce") == 0) resourcetype = XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_TYPE_CE;
                else if (strcasecmp (tmprt, "wn") == 0) resourcetype = XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_TYPE_WN;
                else if (strcasecmp (tmprt, "rb") == 0) resourcetype = XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_TYPE_RB;
                else
                {
                    lcmaps_log(LOG_ERR, "%s: Error: Resource type unknown: \"%s\". Only accepting: \"ce\", \"se\", \"rb\" and \"wn\"\n", logstr, tmprt);
                    return LCMAPS_MOD_FAIL;
                }
            }
            i++;
        }

        /* Action Type - Set the action type to CE, WN, RB, SE */
        else if ( ( (strcasecmp(argv[i], "-actiontype") == 0) ||
                    (strcasecmp(argv[i], "--actiontype") == 0)) &&
		  (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                char * tmprt = argv[i + 1];

                if (strcasecmp (tmprt, "access-file") == 0)      actiontype = XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_TYPE_ACCESS_FILE;
                else if (strcasecmp (tmprt, "queue") == 0)       actiontype = XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_TYPE_QUEUE;
                else if (strcasecmp (tmprt, "execute-now") == 0) actiontype = XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_TYPE_EXEC_NOW;
                else
                {
                    lcmaps_log(LOG_ERR, "%s: Error: Action type unknown: \"%s\". Only accepting: \"queue\", \"execute-now\" and \"access-file\"\n", logstr, tmprt);
                    return LCMAPS_MOD_FAIL;
                }
            }
            i++;
        }
	
        else if ( (strcasecmp(argv[i], "-authorization-only") == 0) ||
                  (strcasecmp(argv[i], "--authorization-only") == 0) )
        {
            pep_set_use_authorization_only(1);
        }
        
	else if ( (strcasecmp(argv[i], "-use-dn-and-fqans") == 0) ||
                  (strcasecmp(argv[i], "--use-dn-and-fqans") == 0) )
        {
            use_x509_chain = 0;
        }

	else if ( (strcasecmp(argv[i], "-send-cert-chain-attribute") == 0) ||
                  (strcasecmp(argv[i], "--send-cert-chain-attribute") == 0) )
        {
            use_cert_chain_attribute = 1;
        }

	else if ( (strcasecmp(argv[i], "-use-dn-from-credential-data") == 0) ||
                  (strcasecmp(argv[i], "--use-dn-from-credential-data") == 0) )
        {
	    use_dn_from_credential_data = 1;
        }

	else if ( (strcasecmp(argv[i], "-proxy-is-unverified") == 0) ||
                  (strcasecmp(argv[i], "--proxy-is-unverified") == 0) )
        {
            cert_client_side_verified = 0;
        }

	else if ( (strcasecmp(argv[i], "-proxy-has-been-verified") == 0) ||
                  (strcasecmp(argv[i], "--proxy-has-been-verified") == 0) )
        {
            cert_client_side_verified = 1;
        }

        else if ( (strcasecmp(argv[i], "-enable-poolindex-fix") == 0) ||
                  (strcasecmp(argv[i], "--enable-poolindex-fix") == 0) )
        {
            use_pool_index_bug_fix = 1;
        }


        /* Override hostname - When evaluating the host certificate, override the expected Hostname to the following setting */
        else if ( ( (strcasecmp(argv[i], "-override-expected-hostname") == 0) ||
                    (strcasecmp(argv[i], "--override-expected-hostname") == 0) ) &&
                  (i + 1 < argc) )
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                /* Clean the mem */
                xacml_io_free_hostname_to_match_client_cert();

                /* Set the expected hostname */
                xacml_io_set_hostname_to_match_client_cert(argv[i + 1]);
                lcmaps_log(LOG_DEBUG, "%s: Setting expected hostname of the service to \"%s\"", logstr, argv[i + 1]);
            }
            i++;
        }

        /* Set network connection timings */
#if 0
        int              realtime_timeout      = 120;      /* seconds */
        int              backoff_milli_seconds = 200;     /* milli seconds */
        int              socket_timeout        = 30000;   /* milli seconds */
#endif

        else if ( ( (strcasecmp(argv[i], "-socket-timeout") == 0) ||
                    (strcasecmp(argv[i], "--socket-timeout") == 0)) &&
                  (i + 1 < argc) )
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
		errno=0;
                long_val = strtol(argv[i + 1], NULL, 10);
		if (errno!=0 || long_val<0)   {
		    lcmaps_log(LOG_ERR,"%s: Cannot convert %s into socket timeout.\n",
			    logstr, argv[i + 1]);
		    return LCMAPS_MOD_FAIL;
		}
                xacml_io_set_socket_timeout(long_val);
            }
            i++;
        }
        else if ( ( (strcasecmp(argv[i], "-connection-timeout") == 0) ||
                    (strcasecmp(argv[i], "--connection-timeout") == 0)) &&
                  (i + 1 < argc) )
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
		errno=0;
                long_val = strtol(argv[i + 1], NULL, 10);
		if (errno!=0 || long_val<0)   {
		    lcmaps_log(LOG_ERR,"%s: Cannot convert %s into connection timeout.\n",
			    logstr, argv[i + 1]);
		    return LCMAPS_MOD_FAIL;
		}
                xacml_io_set_connection_timeout(long_val);
            }
            i++;
        }
        else if ( ( (strcasecmp(argv[i], "-incremental-backoff") == 0) ||
                    (strcasecmp(argv[i], "--incremental-backoff") == 0)) &&
                  (i + 1 < argc) )
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
		errno=0;
		long_val=strtol(argv[i + 1], NULL, 10);
		if (errno!=0 || long_val<0)	{
		    lcmaps_log(LOG_ERR,"%s: Cannot convert %s into incremental_backoff.\n",
			    logstr, argv[i + 1]);
		    return LCMAPS_MOD_FAIL;
		}
                xacml_io_set_incremental_backoff(long_val);
            }
            i++;
        }
        else if ( (strcasecmp(argv[i], "-enable-keepalive") == 0) ||
                  (strcasecmp(argv[i], "--enable-keepalive") == 0) )
        {
            xacml_io_set_keepalive(XACML_KEEPALIVE_ENABLED);
        }
        else if ( (strcasecmp(argv[i], "-disable-keepalive") == 0) ||
                  (strcasecmp(argv[i], "--disable-keepalive") == 0) )
        {
            xacml_io_set_keepalive(XACML_KEEPALIVE_DISABLED);
        }
        else if ( (strcasecmp(argv[i], "-use-system-default-keepalive") == 0) ||
                  (strcasecmp(argv[i], "--use-system-default-keepalive") == 0) )
        {
            xacml_io_set_keepalive(XACML_KEEPALIVE_SYSTEMDEFAULT);
        }
        /* Parameter unknown */
        else
        {
            lcmaps_log(LOG_ERR, "%s: Error in initialization parameter: %s (failure)\n", logstr, argv[i]);
            return LCMAPS_MOD_FAIL;
        }
    }

    /*
     * TODO: Add file open check on both -cert and -key set file paths (existance and such)
     */


    /* Without an endpoint, this plugin is pointless */
    if (n_endpoints == 0)
    {
        lcmaps_log(LOG_ERR, "%s: There are no endpoints configured to contact. Please configure at least one --endpoint <url> option.\n", logstr);
        return LCMAPS_MOD_FAIL;
    }

    /* Default *overall* retrycount, if not set on commandline:
     * - For multiple endpoints set to 1, i.e. each endpoint is tried once,
     * - For one endpoint, set to two, i.e. this one endpoint is tried twice. */
    if (!endpoint_retrycount_is_set)
	endpoint_retrycount = (n_endpoints > 1 ? 1 : 2);

    for (u = 0; u < n_endpoints; u++)
        lcmaps_log(LOG_DEBUG, "%s: Using endpoint: %s\n", logstr, endpoints[u]);

    switch (endpoint_selection_strategy)
    {
        case ENDPOINT_SELECTION_ROUND_ROBIN_TOP_TO_BOTTOM :
            lcmaps_log(LOG_DEBUG, "%s: Endpoint selection strategy set to \"round-robin, starting from the top of the list\"\n", logstr);
            break;
        case ENDPOINT_SELECTION_ROUND_ROBIN_RANDOM_START  :
            lcmaps_log(LOG_DEBUG, "%s: Endpoint selection strategy set to \"round-robin with random starting point in the list\"\n", logstr);
            break;
        case ENDPOINT_SELECTION_RANDOM                    :
            lcmaps_log(LOG_DEBUG, "%s: Endpoint selection strategy set to \"random endpoint selection\"\n", logstr);
            break;
	default:
	    lcmaps_log(LOG_ERR,
		"%s: Internal error, invalid Endpoint selection strategy %d\n",
		logstr,endpoint_selection_strategy);
	    return LCMAPS_MOD_FAIL;
    }


    /*
     *  If the private key and/or the public key (certificate) is not set,
     *  then try to look at the X509_USER_CERT and X509_USER_KEY environment vars for it
     */
    if (!xacml_io_certFilePub)
    {
        if (getenv ("X509_USER_CERT"))
        {
            xacml_io_certFilePub  = strdup(getenv ("X509_USER_CERT"));
            lcmaps_log(LOG_DEBUG,
		    "%s: Using as client SSL credential the certificate from "
		    "$X509_USER_CERT \"%s\".\n", logstr, xacml_io_certFilePub);
        }
    }
    if (!xacml_io_certFilePriv)
    {
        if (getenv ("X509_USER_KEY"))
        {
            xacml_io_certFilePriv  = strdup(getenv ("X509_USER_KEY"));
            lcmaps_log(LOG_DEBUG,
		    "%s: Using as client SSL credential the ""private key from "
		    "$X509_USER_KEY \"%s\".\n", logstr, xacml_io_certFilePriv);
        }
    }


    /*
     * Test if both the certificate file and private key file are set or none to go look for the $X509_USER_PROXY setting
     */
    if ((!xacml_io_certFilePub) && (xacml_io_certFilePriv))
    {
        lcmaps_log(LOG_ERR, "%s: Error: Only -key was set, not -cert. Both or neither must be set. Aborting to avoid confusion and/or unintentional use of the credentials\n", logstr);
        return LCMAPS_MOD_FAIL;
    }

    if ((xacml_io_certFilePub) && (!xacml_io_certFilePriv))
    {
        lcmaps_log(LOG_ERR, "%s: Error: Only -cert was set, not -key. Both or neither must be set. Aborting to avoid confusion and/or unintentional use of the credentials\n", logstr);
        return LCMAPS_MOD_FAIL;
    }


    /* The host credentials should be available here. */
    if (xacml_io_certFilePub && xacml_io_certFilePriv)
    {
        setResourceX509ID = 1;
    }

    /*
     * If the private key and/or the certificate can not be found, then this is
     * a clear initialization failure, unless the $X509_USER_PROXY provides for
     * more information.
     * This will come in the next stage, the plug-in run
     * The SSL connection to the service requires client side credentials
     */
    if ((!xacml_io_certFilePub) || (!xacml_io_certFilePriv))
    {
        lcmaps_log(LOG_DEBUG,
		"%s: Warning: No client SSL credentials provided yet. "
		"Will look for them at the $X509_USER_PROXY in the plug-in "
		"run stage\n", logstr);
        lcmaps_log(LOG_DEBUG,
		"%s:          Checked at \'-cert\', \'-key\', $X509_USER_CERT "
		"and $X509_USER_KEY.\n", logstr);
        /* return LCMAPS_MOD_FAIL; */
    }
    else
    {
        lcmaps_log(LOG_DEBUG,
		"%s: Using the following credentials to contact the service:\n",
		logstr);
        lcmaps_log(LOG_DEBUG,
		"%s:       certificate: \"%s\"\n",logstr,xacml_io_certFilePub);
        lcmaps_log(LOG_DEBUG,
		"%s:       private key: \"%s\"\n",logstr,xacml_io_certFilePriv);
    }

    /* Set certdir from other means if not specified on the cmdline
     * including gLExec->LCMAPS */
    if (!xacml_io_certCAPath)
    {
	char *value=getenv("X509_CERT_DIR");
	xacml_io_certCAPath=value ? strdup(value) : strdup(DEFAULT_CERT_DIR);
	if (xacml_io_certCAPath==NULL)	{
	    lcmaps_log(LOG_ERR,"%s: Out of memory\n",logstr);
	    return LCMAPS_MOD_FAIL;
        }
	lcmaps_log (LOG_INFO, "%s: Using certificate CA path: %s\n",
		logstr, xacml_io_certCAPath);
    }

    lcmaps_log(LOG_DEBUG,
	    "%s: Setting socket timeout to \"%ld\" milliseconds\n",
	    logstr, xacml_io_get_socket_timeout());
    lcmaps_log(LOG_DEBUG,
	    "%s: Setting connection timeout to \"%ld\" seconds\n",
	    logstr, xacml_io_get_connection_timeout());
    lcmaps_log(LOG_DEBUG,
	    "%s: Setting initial incremental backoff to \"%ld\" milliseconds\n",
	    logstr, xacml_io_get_incremental_backoff());
    lcmaps_log(LOG_DEBUG,
	    "%s: Setting TCP keepalive is %s\n",
	    logstr, xacml_io_get_keepalive() ? "enabled" : "disabled");

    lcmaps_log(LOG_DEBUG,"%s: Initialization succeeded\n", logstr);

    return LCMAPS_MOD_SUCCESS;
}





/******************************************************************************
Function:   plugin_introspect
Description:
    return list of required arguments
Parameters:

Returns:
    LCMAPS_MOD_SUCCESS : succes
    LCMAPS_MOD_FAIL    : failure
******************************************************************************/
int plugin_introspect(
        int * argc,
        lcmaps_argument_t ** argv
)
{
    const char * logstr = "lcmaps_plugin_scas-plugin_introspect()";
    static lcmaps_argument_t x509_chain_argList[] = {
        { "user_dn"        , "char *"                ,  1, NULL},
        { "px509_chain"    , "STACK_OF(X509) *"      ,  0, NULL},
        { "voms_data_list" , "lcmaps_vomsdata_t *"   ,  0, NULL},
        { NULL             , NULL                    , -1, NULL}
    };
    static lcmaps_argument_t dn_fqans_argList[] = {
        { "user_dn"        , "char *"                ,  0, NULL},
        { "fqan_list"      , "char **"               ,  0, NULL},
	{ "nfqan"          , "int"                   ,  0, NULL},
        { NULL             , NULL                    , -1, NULL}
    };

    lcmaps_log(LOG_DEBUG,"%s: introspecting\n", logstr);

    *argv = use_x509_chain ? x509_chain_argList : dn_fqans_argList;

    lcmaps_log(LOG_DEBUG,"%s: before lcmaps_cntArgs()\n", logstr);
    *argc = lcmaps_cntArgs(*argv);
    lcmaps_log(LOG_DEBUG,"%s: address first argument: %p\n", logstr,(void*)(*argv));

    lcmaps_log(LOG_DEBUG,"%s: Introspect succeeded\n", logstr);
    return LCMAPS_MOD_SUCCESS;
}


/******************************************************************************
Function:   plugin_run
Description:
    Gather credentials for LCMAPS
Parameters:
    argc: number of arguments
    argv: list of arguments
Returns:
    LCMAPS_MOD_SUCCESS: authorization succeeded
    LCMAPS_MOD_FAIL   : authorization failed
******************************************************************************/
int plugin_run(
        int argc,
        lcmaps_argument_t * argv
)
{
    return plugin_run_or_verify(argc, argv, PLUGIN_RUN);
}


/******************************************************************************
Function:   plugin_verify
Description:
    Verify if user is entitled to use local credentials based on his grid
    credentials. This means that the site should already have been set up
    by, e.g., LCMAPS in a previous run. This method will not try to setup
    account leases, modify (distributed) Unix account information, etc. etc.
    The outcome should be identical to that of plugin_run().
    In this particular case "plugin_verify()" is identical to "plugin_run()"

Parameters:
    argc: number of arguments
    argv: list of arguments
Returns:
    LCMAPS_MOD_SUCCESS: authorization succeeded
    LCMAPS_MOD_FAIL   : authorization failed
******************************************************************************/
int plugin_verify(
        int argc,
        lcmaps_argument_t * argv
)
{
    return plugin_run_or_verify(argc, argv, PLUGIN_VERIFY);
}

static int plugin_run_or_verify(
        int argc,
        lcmaps_argument_t * argv,
        int lcmaps_mode
)
{
/*********** Beg: Initialization *********/
    const char       *  logstr      = "lcmaps_plugin_scas_client_run()";
    /* Default return code is fail */
    int                 plugin_rc = LCMAPS_MOD_FAIL;

    const char       *  encoded_dn = "(void)";

    subject_attr_t	certdata =  {NULL,  /* subject_x509_id */
				     NULL,  /* subject_x509_issuer */
				     NULL,  /* ca_issuer_dn */
				     "",    /* validity_not_before */
				     "",    /* validity_not_after */
				     NULL,  /* certificate_serial_number */
				     NULL,  /* ca_serial_number */
				     0,	    /* numpols */
				     NULL,  /* ca_policy_oid */
				     NULL}; /* cert_chain */
    lcmaps_vomsdata_t * vomsdata  = NULL;
    /* myChain is used when using proxy as input */
    STACK_OF(X509)   *  myChain = NULL;
    /* fake_voms is used (together with user_dn) when using DN+FQANs as input */
    lcmaps_voms_t       fake_voms;
    /* first entry should be &fake_voms, better set that in code itself */
    lcmaps_vomsdata_t   fake_vomsdata={NULL, 1, NULL, NULL};

    /* user_dn is either already input or obtained from certdata. It is after
     * that only used for logging. Same for primary_fqan */
    char             *  user_dn = NULL, *primary_fqan = NULL;
    int                 cnt_dn = 0;
    char	    **	fqans  = NULL, ** user_dn_list = NULL;
    int			nfqans = 0;

    void *value = NULL;

#ifdef RTLD_DEFAULT
    /* Note: don't call it lcmaps_is_set_to_verify_voms_attributes, since that
     * symbol is (probably) already declared in lcmaps_basic.h */
    lcmaps_is_set_to_verify_voms_attributes_t *lcmapsIsSetToVerifyVomsAttributes=NULL;
#endif
    int			voms_client_side_verified=-1;

    /* next will be the resource/dns-host-name attribute */
    char             *  local_host = NULL;
    /* next will be the resource/resource-x509-id attribute */
    char             *  resource_x509_id     = NULL;
    /* next will be the resource/resource-x509-issuer attribute */
    char             *  resource_x509_issuer = NULL;

    int                 i = 0;
    unsigned int        j, u,random_choice = 0, prev_endpoint;

/*********** Beg: SAML2-XACML2 *********/ 
    /* NOTE: request and response are pointers, make sure they're initialized */
    xacml_request_t request=NULL;
    xacml_response_t response=NULL;
    XACML_QUERY_ANSWER_T rc = XACML_QUERY_ALLOWED;
    int should_do_high_level_retry = 0;

/* Hack: SAML2-XACML2 must not write to stdout */
    int pfd[2]              = {-1,-1};
    FILE *f_err             = NULL;
/* Hack: SAML2-XACML2 must not write to stdout */

/*********** End: SAML2-XACML2 *********/
/*********** End: Initialization *********/

    /* Initialize fake_voms and further initialize fake_vomsdata */
    memset(&fake_voms, 0, sizeof(lcmaps_voms_t));
    fake_vomsdata.voms=&fake_voms;
    
    /*
     * The beginning
     */
    if (lcmaps_mode == PLUGIN_RUN)
        logstr = "lcmaps_plugin_scas_client-plugin_run()";
    else if (lcmaps_mode == PLUGIN_VERIFY)
        logstr = "lcmaps_plugin_scas_client-plugin_verify()";
    else
    {
        lcmaps_log(LOG_ERR,
            "lcmaps_plugin_scas_client-plugin_run_or_verify(): "
	    "attempt to run plugin in invalid mode: %d\n",
            lcmaps_mode);
        goto cleanup; /* cleanup and fail */
    }

    /*
     *  If the private key and/or the public key (certificate) is not set,
     *  then try the X509_USER_PROXY environment variable
     */
    if ((!xacml_io_certFilePub) || (!xacml_io_certFilePriv)) {
        char * x509_user_proxy = NULL;

        if (!(x509_user_proxy = getenv ("X509_USER_PROXY"))) {
            lcmaps_log(LOG_ERR,
		    "%s: No client side credentials for SSL handshake to the "
		    "SCAS service presented. Not in the configuration file "
		    "and not in the $X509_USER_{PROXY,CERT,KEY} environment "
		    "variables.\n", logstr);
            goto cleanup; /* cleanup and fail */
        } else {
            lcmaps_log(LOG_DEBUG,
		    "%s: Using $X509_USER_PROXY \"%s\"\n",
		    logstr, x509_user_proxy);

            xacml_io_certFilePub  = strdup(x509_user_proxy);
            xacml_io_certFilePriv = strdup(x509_user_proxy);
        }
    }




    /***  Acquiring  information  ***/

    lcmaps_log(LOG_DEBUG,
	    "%s: Acquiring information from LCMAPS framework\n", logstr);

    if (use_x509_chain) { /* Use proxy chain is input for subject credentials */
	/* First get chain from LCMAPS */
	if ( (value = lcmaps_getArgValue("px509_chain", "STACK_OF(X509) *",
			                 argc, argv)) ==  NULL ||
	     (myChain = *(STACK_OF(X509) **) value) == NULL )
	{
	    lcmaps_log(LOG_ERR, "%s: no X.509 chain is set.\n", logstr);
	    goto cleanup; /* cleanup and fail */
	}
	
	lcmaps_log(LOG_DEBUG, "%s: found X.509 chain.\n", logstr);

	/* get VOMS from LCMAPS */
	value = lcmaps_getArgValue("voms_data_list", "lcmaps_vomsdata_t *",
				   argc, argv);

	if ( value && (vomsdata = *(lcmaps_vomsdata_t **) value) ) {
	    lcmaps_log(LOG_DEBUG,"%s: found VOMS data\n", logstr);
	    for (i=0; i<vomsdata->nvoms; i++)
		lcmaps_log(LOG_DEBUG,
			"%s: vomsdata->voms[%d].voname = %s\n",
			logstr, i, vomsdata->voms[i].voname);
	} else {
	    lcmaps_log(LOG_INFO,
		"%s: No VOMS AC(s) found by the framework in the proxy chain.\n",
		logstr);
	}

	/* Check whether we or server should to the VOMS verification */

#ifdef RTLD_DEFAULT /* Most UNIX now support RTLD_DEFAULT (POSIX reserved) */
	/* Try to get the lcmaps_is_set_to_verify_voms_attributes symbol */
	dlerror();
	lcmapsIsSetToVerifyVomsAttributes=(lcmaps_is_set_to_verify_voms_attributes_t*)dlsym(RTLD_DEFAULT,"lcmaps_is_set_to_verify_voms_attributes");
	if (dlerror()==NULL)	{
	    /* Symbol is found, use it */
	    voms_client_side_verified=lcmapsIsSetToVerifyVomsAttributes();
	} else	{
	    /* We either use an ancient LCMAPS version (<1.4.27) or we are using
	     * the without-gsi library. In the latter case, we cannot not be
	     * here (should be in the NOT use_x509_chain code). Best to set to
	     * undefined. */
	    lcmaps_log(LOG_WARNING,
		"%s: Cannot determine whether VOMS is verified: "
		"Cannot find lcmaps_is_set_to_verify_voms_attributes() symbol.\n",
		logstr);
	    voms_client_side_verified=-1;
	}
#else	/* No RTLD_DEFAULT: we cannot assume we have the symbol: that would be
	   problem when running with the -without-gsi mode, but in that case, we
	   should not be here in any case (not using x509_chain), so setting to
	   undefined is safe */
	lcmaps_log(LOG_WARNING,
		"%s: Cannot determine whether VOMS is verified: "
		"no RTLD_DEFAULT support.\n", logstr);
	voms_client_side_verified=-1;
#endif

	/* Fill the subject attribute structure */
	if (fill_subject_attributes(&certdata, myChain)<0)  {
	    lcmaps_log(LOG_ERR,
		    "%s: fill_subject_attributes() failed\n",logstr);
	    goto cleanup; /* cleanup and fail */
	}

	/* First try to obtain DN from the credential data (i.e. stored by other
	 * plugins, we normally don't use this. */
	user_dn_list = getCredentialData(DN, &cnt_dn);
	if (cnt_dn>0)   {
	    /* Log already registered DNs */
	    for (i=0; i<cnt_dn; i++)
		lcmaps_log(LOG_DEBUG,"%s: found registered DN[%d/%d]: %s\n",
			logstr, i+1, cnt_dn, user_dn_list[i]);

	    /* Do we want to use the DN registered by other plugins? */
	    if (use_dn_from_credential_data)	{
		lcmaps_log(LOG_INFO,
			"%s: using DN from credential data (instead of from certificate): %s\n",
			logstr, user_dn_list[0]);
		/* allocated using X509_NAME_oneline() */
		free(certdata.subject_x509_id);
		certdata.subject_x509_id=strdup(user_dn_list[0]);
		if (certdata.subject_x509_id==NULL) {
		    lcmaps_log(LOG_ERR, "%s: out-of-memory\n", logstr);
		    goto cleanup;
		}
		if (cert_client_side_verified==1) {
		    /* Value is already set with cmdline option, don't override it */
		    lcmaps_log(LOG_INFO,
			    "%s: assuming DN is verified, although using that from credential data\n",
			    logstr);
		}
	    }
	
	    /* Set user_dn, will be used for logging by xacml_query_scas */
	    user_dn=certdata.subject_x509_id;
	} else { /* No credentials registered yet */
	    /* Set user_dn, will be used for logging by xacml_query_scas */
	    user_dn=certdata.subject_x509_id;

	    /* Register DN in LCMAPS when found and not yet set */
	    /* Fix - Brian: Nov 12 2010 */
	    /* addCredentialData creates an internal copy of user_dn */
	    addCredentialData (DN, &user_dn);
	}

	/* Set primary_fqan, will be used for logging by xacml_query_scas */
	primary_fqan=(vomsdata && vomsdata->nvoms>0
			    ? vomsdata->voms[0].fqan_unix[0].fqan
			    : "");

    } else {	/* i.e. not use_x509_chain */
	/* Don't have chain, try subject DN (LCMAPS does not provide us with
	 * issuer DN or other cert info */

	/* First try to obtain DN from the credential data (i.e. stored by other
	 * plugins */
	user_dn_list = getCredentialData(DN, &cnt_dn);
	if (cnt_dn>0)	{
	    for (i=0; i<cnt_dn; i++)
		lcmaps_log(LOG_DEBUG,"%s: found registered DN[%d/%d]: %s\n",
			logstr, i+1, cnt_dn, user_dn_list[i]);
	    user_dn=user_dn_list[0];
	    fqans = getCredentialData(LCMAPS_VO_CRED_STRING, &nfqans);
	    lcmaps_log(LOG_DEBUG,"%s: found %d FQANs\n", logstr, nfqans);
	} else {
	    /* Try plugin arguments (put in at plugin_initialize stage) */
	    if ( (value = lcmaps_getArgValue("user_dn", "char *", argc, argv)) == NULL ||
		 (user_dn = *(char **)value) == NULL )
	    {
		lcmaps_log(LOG_ERR, "%s: no subject DN is set.\n", logstr);
		goto cleanup; /* cleanup and fail */
	    }
	    lcmaps_log(LOG_DEBUG,"%s: found DN: %s\n", logstr, user_dn);
	    /* We don't have a DN registered in LCMAPS or we would have found it
	     * above, so push it to the end-result registry now */
	    addCredentialData(DN, &user_dn);
	    /* Try to obtain the FQANs */
	    if ( (value = lcmaps_getArgValue("nfqan", "int", argc, argv) ) )	{
		nfqans = *(int *)value;
		lcmaps_log(LOG_DEBUG, "%s: found nfqan: %d\n", logstr, nfqans);
		if (nfqans>0)	{
		    lcmaps_log(LOG_DEBUG,
			    "%s: the list of FQANs should contain %d elements\n",
			    logstr, nfqans);
		    if ( (value = lcmaps_getArgValue("fqan_list", "char **",
						     argc, argv)) )
			fqans = *(char ***)value;
		}
	    }
	}

	/* provided input is just DN and FQANs, most likely not verified. */
	if (cert_client_side_verified==-1)  {
	    /* Value wasn't know yet, set it now */
	    cert_client_side_verified=0;
	} else if (cert_client_side_verified==1) {
	    /* Value is already set with cmdline option, don't override it */
	    lcmaps_log(LOG_INFO,
		    "%s: assuming proxy is verified, although using only DN(+FQANs)\n",
		    logstr);
	}
	voms_client_side_verified=0;

	/* Set certdata entry */
	certdata.subject_x509_id=user_dn;

	/* Fill `fake' vomsdata structure */
	fake_voms.nfqan=nfqans;
	if (nfqans>0)	{
	    fake_voms.fqan_unix=(lcmaps_fqan_unix_t*)malloc((size_t)nfqans*sizeof(lcmaps_fqan_unix_t));
	    if (fake_voms.fqan_unix==NULL)	{
		lcmaps_log(LOG_ERR, "%s: out of memory\n", logstr);
		goto cleanup; /* cleanup and fail */
	    }
	    for (i=0; i<nfqans; i++)
		fake_voms.fqan_unix[i].fqan=fqans[i];

	    /* Set vomsdata: only when we have a valid number of FQANs */
	    vomsdata=&fake_vomsdata;

	    /* Set primary_fqan for logging by xacmlqueryscas() */
	    primary_fqan=fqans[0];

	}
    }	/* end of not use_x509_chain */

    /* Work goes here */
    lcmaps_log(LOG_DEBUG, "%s: Starting XACML framework\n", logstr);

    xacml_init();
    xacml_request_init(&request);

    xacml_request_set_io_descriptor(request, &xacml_io_ssl_descriptor);
    xacml_response_init(&response);

    lcmaps_log(LOG_DEBUG,
	    "%s: XACML framework initialized, grabbing information\n", logstr);

    /************ B: Construction of the XACML request ***********/


    if (pep_construct_request_subject (request,
		       &certdata,		    /* certificate chain */
		       vomsdata,		    /* VOMS data */
		       cert_client_side_verified,   /* proxy is verified? */
		       voms_client_side_verified,   /* VOMS is verified? */
		       use_dn_from_credential_data  /* DN from cert or creddata? */
                                      ))
    {
        lcmaps_log(LOG_ERR,
		"%s: XACML: Error in constructing request subject\n", logstr);
        goto cleanup; /* cleanup and fail */
    }

    /* Construct action */
    if (pep_construct_request_action (request,
                                      actiontype
                                     ))
    {
        lcmaps_log(LOG_ERR,
		"%s: XACML: Error in constructing request action\n", logstr);
        goto cleanup; /* cleanup and fail */
    }




    lcmaps_log(LOG_DEBUG, "%s: Fetching DNS/Host name\n", logstr);
#ifndef MAXHOSTNAMELEN
   #define MAXHOSTNAMELEN ((size_t)255)
#endif /* MAXHOSTNAMELEN */
    local_host = (char *) malloc (sizeof(char) * MAXHOSTNAMELEN);
    gethostname (local_host, MAXHOSTNAMELEN);
    lcmaps_log(LOG_DEBUG, "%s: DNS/Host name is: %s\n", logstr, local_host);


    /* Add resource-x509-id and resource-x509-issuer from correct files: either
     * proxy or hostcert/hostkey */
    if (setResourceX509ID)
    {
        FILE *              lcmaps_pem_fh = NULL;
        X509 *              px509 = NULL;
        #define             PEM_BUF_SIZE ((size_t)10000)
        char                pem_buf[PEM_BUF_SIZE];
        size_t              nbytes = 0;

        memset (pem_buf, '\0', PEM_BUF_SIZE);

        /* Read PEM string */
        if ( (lcmaps_pem_fh = fopen(xacml_io_certFilePub, "r")) == NULL)
        {
            lcmaps_log (LOG_WARNING,
		    "Could not open file for reading: %s\n",
		    xacml_io_certFilePub);
        }
        else
        {
            nbytes = fread(pem_buf, sizeof(pem_buf), (size_t)1, lcmaps_pem_fh);
            if (feof(lcmaps_pem_fh))
            {
                lcmaps_log(LOG_DEBUG,
			"%s: EOF reached: %s\n", logstr, xacml_io_certFilePub);
            }

            lcmaps_log(LOG_DEBUG,
		    "%s: nbytes = %lu. Error = %d\n",
		    logstr, (long unsigned)nbytes, ferror(lcmaps_pem_fh));
            fclose(lcmaps_pem_fh);

            if (pem_string_to_x509(&px509, pem_buf) != 0)
            {
                lcmaps_log (LOG_ERR,
			"%s: Error in pem_string_to_x509()\n", logstr);
            }
            else
            {
                resource_x509_id     = X509_NAME_oneline(X509_get_subject_name(px509), NULL, 0);
                resource_x509_issuer = X509_NAME_oneline(X509_get_issuer_name(px509), NULL, 0);

                /* Mandatory clean up - Fix by Brian */
                X509_free(px509);
                px509 = NULL;
            }
        }

    }


    /* Construct resource */
    if (pep_construct_request_resource (request,
                                        resourcetype,
                                        local_host,
                                        NULL,
                                        resource_x509_id,
                                        resource_x509_issuer
                                       ))
    {
        lcmaps_log(LOG_ERR,
		"%s: XACML: Error in constructing request resource\n", logstr);
        free (local_host);
        local_host = NULL;
        goto cleanup; /* cleanup and fail */
    }

    if (local_host)
    {
        free (local_host);
        local_host = NULL;
    }
    if (resource_x509_id)
    {
        free (resource_x509_id);
        resource_x509_id = NULL;
    }
    if (resource_x509_issuer)
    {
        free (resource_x509_issuer);
        resource_x509_issuer = NULL;
    }
    /************ E: Construction of the XACML request ***********/


    lcmaps_log(LOG_DEBUG,
	    "%s: XACML framework: Registering Obligation Handlers\n", logstr);


    /*
     * TODO Test: Explicitly ignore the Obligation Handler
     *            by not even loading/registering the obligation handler.
     */
    if (pep_get_use_authorization_only() == 0)
    {
        /************ B: Register Obligation Handlers *********/
        if (pep_register_obligation_handlers (request))
        {
            lcmaps_log(LOG_ERR,
		    "%s: XACML: Error in Registering Obligation Handlers\n",
		    logstr);
            goto cleanup; /* cleanup and fail */
        }
        /************ E: Register Obligation Handlers *********/
    }


    if (xacml_io_certFilePub == NULL || xacml_io_certFilePriv == NULL)
    {
        lcmaps_log(LOG_ERR,
		"%s: XACML: Error can't work without a certificate for the "
		"connection to the service\n", logstr);
        goto cleanup; /* cleanup and fail */
    }


    /************* B: Silence the stdout and stderr of the SAML2-XACML2 library *************/
    /* This used to silence stdout and stderr. Now we open a pipe and let the
     * S2X2-C-lib log to that. Afterwards log the buffered output to lcmaps log.
     * We also now use the new xacml lib functionality to provide a stream for
     * the logging instead of using stderr.
     */
    if (pipe(pfd)) {
        lcmaps_log(LOG_ERR,
	    "%s: XACML: Error - cannot open new pipes: (errno=%d, %s)\n",
	    logstr, errno, strerror(errno));
        goto cleanup; /* cleanup and fail */
    }
    if ( (fcntl(pfd[1], F_SETFD, FD_CLOEXEC) == -1) ||
	 (fcntl(pfd[0], F_SETFD, FD_CLOEXEC) == -1) ) {
        lcmaps_log(LOG_ERR,
	    "%s: XACML: Error - cannot set pipes to CLOEXEC: (errno=%d, %s)\n",
	    logstr, errno, strerror(errno));
        goto cleanup; /* cleanup and fail */
    }
    if ( (fcntl(pfd[1], F_SETFL, O_NONBLOCK) == -1) ||
	 (fcntl(pfd[0], F_SETFL, O_NONBLOCK) == -1) ) {
        lcmaps_log(LOG_ERR,
	    "%s: XACML: Error - cannot set write pipe to non-blocking: (errno=%d, %s)\n",
	    logstr, errno, strerror(errno));
        goto cleanup; /* cleanup and fail */
    }

    /* Open stream belonging to write-pipe */
    if ( (f_err=fdopen(pfd[1], "w"))==NULL )  {
	lcmaps_log(LOG_ERR,
	    "%s: XACML: Error - cannot get stream for pipe: (errno=%d, %s)\n",
	    logstr, errno, strerror(errno));
        goto cleanup; /* cleanup and fail */
    }

    /************* E: Silence the stdout and stderr of the SAML2-XACML2 library *************/

    /* initialize whether we need to retry */
    should_do_high_level_retry = 1;

    /* Set previous endpoint to a never matching value */
    prev_endpoint=n_endpoints;

    /* Execute endpoint selection strategy */
    switch (endpoint_selection_strategy)
    {
        /* Round Robin - Top to Bottom */
        case ENDPOINT_SELECTION_ROUND_ROBIN_TOP_TO_BOTTOM :
            lcmaps_log(LOG_DEBUG,
		    "%s: Starting Round-Robin (top to bottom) strategy\n",
		    logstr);

            for (j = 0; j < endpoint_retrycount && should_do_high_level_retry; j++)
            {
		for (u = 0; u < n_endpoints && should_do_high_level_retry; u++)
                {
		    /* Wait for 50 milli seconds, plus a random number that
		     * randomizes below the 10 milli seconds if we use the same
		     * endpoint */
		    if (u==prev_endpoint)
			usleep (50000 + xacml_io_rand() % 10000);

                    /* Interact with the SCAS */
                    rc = xacmlqueryscas(endpoints[u], request, response,
					f_err, pfd[0],
					&should_do_high_level_retry,
					user_dn, primary_fqan, j+1);

		    /* Renew prev_endpoint */
		    prev_endpoint=u;
                }
            }
            break;
        case ENDPOINT_SELECTION_ROUND_ROBIN_RANDOM_START :
            lcmaps_log(LOG_DEBUG,
		    "%s: Starting Round-Robin (random start) strategy\n",
		    logstr);
            
	    /* Pick a random starting point */
            random_choice = xacml_io_rand() % n_endpoints;

            for (j = 0; j < endpoint_retrycount && should_do_high_level_retry; j++)
            {
		for (u = 0; u < n_endpoints && should_do_high_level_retry; u++)
                {
		    unsigned int selected_endpoint = (random_choice + u) % n_endpoints;

		    /* Wait for 50 milli seconds, plus a random number that
		     * randomizes below the 10 milli seconds if we use the same
		     * endpoint */
		    if (selected_endpoint==prev_endpoint)
			usleep (50000 + xacml_io_rand() % 10000);

                    /* Interact with the SCAS */
                    rc = xacmlqueryscas(endpoints[selected_endpoint], request, response,
					f_err, pfd[0],
					&should_do_high_level_retry,
					user_dn, primary_fqan, j+1);

		    /* Renew prev_endpoint */
		    prev_endpoint=selected_endpoint;
                }
            }
            break;
        case ENDPOINT_SELECTION_RANDOM :
            lcmaps_log(LOG_DEBUG,
		    "%s: Starting Random selection strategy\n", logstr);

            for (j = 0; j < endpoint_retrycount && should_do_high_level_retry; j++)
            {
		for (u = 0; u < n_endpoints && should_do_high_level_retry; u++)
                {
		    /* Pick a random choice */
		    random_choice = xacml_io_rand() % n_endpoints;

		    /* Wait for 50 milli seconds, plus a random number that
		     * randomizes below the 10 milli seconds if we use the same
		     * endpoint */
		    if (random_choice==prev_endpoint)
			usleep (50000 + xacml_io_rand() % 10000);

                    /* Interact with the SCAS */
                    rc = xacmlqueryscas(endpoints[random_choice], request, response,
					f_err, pfd[0],
					&should_do_high_level_retry,
					user_dn, primary_fqan, j+1);

		    /* Renew prev_endpoint */
		    prev_endpoint=random_choice;
                }
            }
            break;
	default:
	    lcmaps_log(LOG_ERR,
		"%s: Internal error, invalid Endpoint selection strategy %d\n",
		logstr,endpoint_selection_strategy);
	    return LCMAPS_MOD_FAIL;
    }

    /* All else is an error, besides the XACML_QUERY_ALLOWD */
    if (rc != XACML_QUERY_ALLOWED)
	goto cleanup; /* cleanup and fail */

    /* Now we can free the overridden hostname when set */
    xacml_io_free_hostname_to_match_client_cert();

    /* This fix is to adjust for LCMAPS version 1.4.6 and older */
    if (use_pool_index_bug_fix)
    {
        lcmaps_log(LOG_DEBUG,
		"%s: Adding POOL_INDEX :  %s\n", logstr, encoded_dn);
        addCredentialData(POOL_INDEX, &encoded_dn);
    }



/* Succesful clean up */
    lcmaps_log(LOG_DEBUG, "%s: Destroying: Obligation handlers\n", logstr);

/*
    xacml_obligation_destroy(UIDGID_handler);
    xacml_obligation_destroy(SecondaryGIDs_handler);
*/

/*
 default_handler
 local_user_name_handler
 UIDGID_handler
 SecondaryGIDs_handler
 AFSToken_handler
*/

    /* Set return value */
    plugin_rc = LCMAPS_MOD_SUCCESS;

cleanup:
    /* Cleanup */

    /* Close the pipe used for the XACML logging */
    /* Read end */
    if (pfd[0] >= 0)	{
	close(pfd[0]);
	pfd[0]=-1;
    }
    /* Write end */
    if (pfd[1] >= 0)	{
	if (f_err)  {  /* Use associated stream */
	    fclose(f_err);
	    f_err=NULL;
	}
	else	{	    /* Use underlying file descriptor */
	    close(pfd[1]);
	}
	pfd[1]=-1;
    }

    lcmaps_log(LOG_DEBUG, "%s: Destroying: response\n", logstr);
    xacml_response_destroy(response);
    response=NULL;  /* set to NULL to prevent double free */

    lcmaps_log(LOG_DEBUG, "%s: Destroying: request\n", logstr);
    xacml_request_destroy(request);
    request=NULL;   /* set to NULL to prevent double free */

    /* Now we can free the overridden hostname when set */
    xacml_io_free_hostname_to_match_client_cert();

    if (use_x509_chain)   {
	free_subject_attributes(&certdata);
    } else {
	free(fake_voms.fqan_unix); fake_voms.fqan_unix=NULL;
	fake_voms.nfqan=0;
    }

    free(local_host); local_host = NULL;
    free(resource_x509_id); resource_x509_id = NULL;
    free(resource_x509_issuer); resource_x509_issuer = NULL;

    lcmaps_log(LOG_INFO,"%s: scas client plugin %s\n",
	    logstr, plugin_rc==LCMAPS_MOD_SUCCESS ? "succeeded" : "failed");

    return plugin_rc;
}



/******************************************************************************
Function:   plugin_terminate
Description:
    Terminate plugin
Parameters:

Returns:
    LCMAPS_MOD_SUCCESS : succes
    LCMAPS_MOD_FAIL    : failure
******************************************************************************/
int plugin_terminate(void)
{
    const char * logstr = "lcmaps_plugin_scas_client-plugin_terminate()";
    lcmaps_log(LOG_DEBUG,"%s: terminating\n", logstr);

    free(xacml_io_certFilePub);  xacml_io_certFilePub  = NULL;
    free(xacml_io_certFilePriv); xacml_io_certFilePriv = NULL;
    free(xacml_io_certCAPath);   xacml_io_certCAPath   = NULL;
    free(xacml_io_certOwner);    xacml_io_certOwner    = NULL;

    return LCMAPS_MOD_SUCCESS;
}





/******************************************************************************
Function:       pem_string_to_x509
Description:    Reads a X509 certificate from a PEM-encoded string
Parameters:
                px: pointer to a X509 certificate
                cert_string: PEM-encoded string
Returns:        0 on success, non-zero otherwise
******************************************************************************/
static int pem_string_to_x509(X509 **px, char *certstring)
{

    BIO *certbio = NULL;

    *px = X509_new();
    if (*px == NULL) return -1;
    /* clean it first */
    X509_free(*px);

    certbio = BIO_new_mem_buf(certstring, -1);
    if (certbio)
    {
        /* This call allocates new X509 struct, so no X509_new() needed */
        *px = PEM_read_bio_X509(certbio, NULL, 0, NULL);
    }

    BIO_free(certbio);
    return 0;
}

/******************************************************************************
Function:       log_fake_stderr
Description:    Dumps the contents of the read pipe to the logfile
Parameters:
                level - log level
                fd - file descriptor to read from
Returns:        void
******************************************************************************/
static void log_fake_stderr(int level, int fd)
{
    char pipe_buf[PIPE_BUFFER_SIZE];
    char *buf,*pos;
    int line;
    ssize_t ret;
    while ((ret = read(fd, pipe_buf, PIPE_BUFFER_SIZE-1)) != 0)
    {
        if (ret == -1)
        {
	    if (errno == EINTR) /* interrupted, try again */
		continue;
	    else if (errno == EAGAIN) /* Nothing there, quit */
		break;
	    else { /* error */
		lcmaps_log(level,
			"%s: Error reading from error buffer: %s\n",
			__func__, strerror(errno));
		return;
	    }
        }

        pipe_buf[ret] = '\0';

	/* Break up buffer on newlines, log each line separately */
	buf=pipe_buf; line=0;

	while (buf[0]!='\0')  {
	    /* Look for next newline */
	    if ( (pos=strchr(buf,'\n')) )
		pos[0]='\0';

	    /* Inline second and onwards lines */
	    if (line) /* indent 2nd- lines */
		lcmaps_log(level, "    %s\n", buf);
	    else
		lcmaps_log(level, "%s\n", buf);

	    /* when a newline was found, continue */
	    if (pos)	{
		buf=&(pos[1]);
		line++;
	    } else
		break;
	}
    }
}

/******************************************************************************
Function:       xacmlqueryscas
Description:    Queries the SCAS service for a authorization decision and mapping
Parameters:
                endpoint
                request
                response
		f_err open file pointer to send log to.
Returns:        XACML_QUERY_ALLOWED on success, non-zero otherwise
******************************************************************************/
static XACML_QUERY_ANSWER_T xacmlqueryscas (const char *endpoint, /* scas */
				 xacml_request_t  request,  /* input */
				 xacml_response_t response, /* output */
				 FILE            *f_err,    /* err stream for soap */
				 int              filedesc, /* pipe for LCMAPS */
				 int             *retry,    /* whether to retry */
				 const char      *user_dn,  /* user DN */
				 const char      *pFQAN,    /* prim. FQAN */
				 unsigned int     try)	    /* which try */
{
    const char * logstr = "lcmaps_plugin_scas_client-xacmlqueryscas()";
    saml_status_code_t   code;
    xacml_decision_t     decision;
    int                  rc_xacml = 0;
    XACML_QUERY_ANSWER_T returncode = XACML_QUERY_ALLOWED;
#ifdef HAVE_XACML_REQUEST_GET_ID
    const char          *req_issuer=NULL, *resp_issuer=NULL;
    const char          *req_id=NULL, *resp_id=NULL, *resp_req_id=NULL;
    char                 issue_time_str[TIME_BUF_LEN]="";
    time_t               issue_time=0;
#else
    char                 *req_issuer= NULL, *resp_issuer=NULL;
#endif	/* HAVE_XACML_REQUEST_GET_ID */

    /* Basic input check */
    if (endpoint == NULL || request == NULL || response == NULL ||
	f_err == NULL || retry == NULL )
    {
	/* If one of these fails, something is very wrong */
	lcmaps_log(LOG_ERR,
		"%s: internal error: one of the parameters is NULL\n", logstr);
	if (retry)
	    *retry=0;
	return XACML_QUERY_PROTOCOL_ERROR;
    }

    /* Log the outgoing request */

    /* note that the issuer will be set to the subject of the request and the
     * time will be determined locally in xacml_query_file() */
    xacml_request_get_subject(request, &req_issuer);
#ifdef HAVE_XACML_REQUEST_GET_ID
    /* Create and set the request ID */
    req_id = xacml_request_set_id(request, NULL);
    lcmaps_log(LOG_INFO,
		"%s: Outgoing request: %s (endpoint \"%s\", subject \"%s\")\n",
		logstr, req_id ? req_id : "unknown", endpoint, req_issuer);
#else
    lcmaps_log(LOG_INFO,
		"%s: Outgoing request to endpoint \"%s\" from \"%s\"\n",
		logstr, endpoint, req_issuer);
#endif /* HAVE_XACML_REQUEST_GET_ID */

    /* Query XACML endpoint */
    rc_xacml = xacml_query_file(endpoint, request, response, f_err);
    fflush(f_err);

    /* Only log the response if we actually have one, can see from the return
     * value. Also log for failed obligations. */
    if (rc_xacml==XACML_RESULT_SUCCESS ||
	rc_xacml==XACML_RESULT_OBLIGATION_FAILED)
    {
	xacml_response_get_issuer(response, &resp_issuer);
#ifdef HAVE_XACML_REQUEST_GET_ID
	xacml_response_get_id(response, &resp_id);
	/* Obtain request ID explicitly from the response, even though it should
	 * match the value above... */
	xacml_response_get_in_response_to(response, &resp_req_id);
	xacml_response_get_issue_instant(response, &issue_time);
	if (issue_time>0)
	    lcmaps_scas_client_time_t_to_string(issue_time,
						issue_time_str,
						TIME_BUF_LEN);
	/* Basic check: is the InResponseTo the same as the request ID */
	if (req_id && resp_req_id && strcmp(req_id, resp_req_id)!=0)
	    lcmaps_log(LOG_WARNING,
		    "%s: warning: response is in in response to \"%s\", "
		    "while request was \"%s\"",
		    logstr, req_id, resp_req_id);
	/* Log the response */
	lcmaps_log(LOG_INFO,
		"%s: Incoming response: %s "
		"(request \"%s\", issuer \"%s\", issue-time %s)\n",
		logstr, resp_id ? resp_id : "unknown",
		req_id ? req_id : "unknown",
		resp_issuer ? resp_issuer : "unknown",
		strlen(issue_time_str) ? issue_time_str : "unknown");
#else
	lcmaps_log(LOG_INFO,
		"%s: Incoming response from issuer \"%s\"\n",
		logstr, resp_issuer ? resp_issuer : "unknown");
#endif	/* HAVE_XACML_REQUEST_GET_ID */

	lcmaps_log(LOG_DEBUG, "%s: XACML: Success with endpoint: \"%s\"\n",
		logstr, endpoint);

	xacml_response_get_saml_status_code(response, &code);
	xacml_response_get_xacml_decision(response, &decision);

	lcmaps_log(LOG_DEBUG,
		"%s: Server \"%s\" said: \"%s\" with decision: \"%s\"\n",
		logstr,
		endpoint,
		saml_status_code_strings[code],
		decision_str(decision));

	/* Determine the exact decision */
	switch (decision)
	{
	    case XACML_DECISION_Permit        :
		/* Specific fix to be compatible with service like SAZ that
		 * do not give out obligations - this must be explicitly
		 * activated */
		if (pep_get_use_authorization_only()) {
		    lcmaps_log(LOG_DEBUG,
			"%s: XACML: Authorized by service \"%s\"\n",
			logstr, endpoint);
		    returncode=XACML_QUERY_ALLOWED;
		    break; /* from switch decision */
		}

		/* Get and parse resulting account information */
		if (parse_and_store_creddata(endpoint)==-1)
		    returncode=XACML_QUERY_PROTOCOL_ERROR;
		else
		    returncode=XACML_QUERY_ALLOWED;
		/* Cleanup the returned and stored list of uids and gids */
		clean_uidgid_lists();
		break; /* from switch decision */
	    case XACML_DECISION_Deny          :
		lcmaps_log (LOG_NOTICE,
		    "%s: XACML: Access denied at endpoint: \"%s\"\n",
		    logstr, endpoint);
		returncode=XACML_QUERY_BANNED;
		break; /* from switch decision */
	    case XACML_DECISION_Indeterminate :
	    case XACML_DECISION_NotApplicable :
	    default:
		lcmaps_log (LOG_ERR,
		    "%s: XACML: Failure: decision \"%s\" at endpoint \"%s\".\n",
		    logstr, decision_str(decision), endpoint);
		returncode=XACML_QUERY_PROTOCOL_ERROR;
		break; /* from switch decision */
	} /* switch decision */
        
	if (rc_xacml==XACML_RESULT_OBLIGATION_FAILED)	{
            lcmaps_log(LOG_ERR,
		"%s: XACML: Obligation failed. The returned"
		" obligation(s) couldn't be processed. Endpoint \"%s\"\n",
		logstr, endpoint);
            returncode=XACML_QUERY_PROTOCOL_ERROR;
	}
    } else {
	/* Top level switch for failures other than obligations: result of
	 * xacml_query_file() */
	switch (rc_xacml)
	{
	    case XACML_RESULT_INVALID_PARAMETER :
		lcmaps_log(LOG_ERR,
		    "%s: XACML: Interaction failed: Invalid parameter. "
		    "Endpoint is: \"%s\"\n", logstr, endpoint);
		returncode=XACML_QUERY_PROTOCOL_ERROR;
		break; /* from switch rc_xacml */
	    case XACML_RESULT_SOAP_ERROR :
		lcmaps_log(LOG_ERR,
		    "%s: XACML: Interaction failed: "
		    "TCP/IP, SSL or SOAP Error with endpoint: \"%s\"\n",
		    logstr, endpoint);
		returncode=XACML_QUERY_PROTOCOL_ERROR;
		break; /* from switch rc_xacml */
	    case XACML_RESULT_INVALID_STATE :
		lcmaps_log(LOG_ERR,
		    "%s: XACML: Interaction failed: Invalid state. "
		    "Endpoint: \"%s\"\n", logstr, endpoint);
		returncode=XACML_QUERY_PROTOCOL_ERROR;
		break; /* from switch rc_xacml */
	    default :
		/* unknown return value from xacml_query_file() */
		lcmaps_log(LOG_ERR,
		    "%s: XACML: Interaction failed: Unknown error. "
		    "Endpoint: \"%s\"\n", logstr, endpoint);
		returncode=XACML_QUERY_PROTOCOL_ERROR;
		break; /* from switch rc_xacml */
	} /* switch rc_xacml */
    }

    /* Are we allowed, denied, or have an error:
     * log and determine if we should retry */
    switch (returncode)
    {
	case XACML_QUERY_ALLOWED        :
	    log_fake_stderr(LOG_NOTICE, filedesc);
	    if (pFQAN)
		lcmaps_log (LOG_NOTICE,
		    "%s: User \"%s\" with first FQAN \"%s\" Permitted"
		    " at endpoint \"%s\"\n",
		    logstr, user_dn, pFQAN, endpoint);
	    else
		lcmaps_log (LOG_NOTICE,
		    "%s: User \"%s\" (no valid FQANs) Permitted"
		    " at endpoint \"%s\"\n",
		    logstr, user_dn, endpoint);
	    /* Do not retry: we have a positive answer */
	    *retry = 0;
	    break; /* switch returncode */
	case XACML_QUERY_BANNED         :
	    log_fake_stderr(LOG_WARNING, filedesc);
	    if (pFQAN)
		lcmaps_log (LOG_WARNING,
		    "%s: User \"%s\" with first FQAN \"%s\" Banned"
		    " at endpoint \"%s\"\n",
		    logstr, user_dn, pFQAN, endpoint);
	    else
		lcmaps_log (LOG_WARNING,
		    "%s: User \"%s\" (no valid FQANs) Banned"
		    " at endpoint \"%s\"\n",
		    logstr, user_dn, endpoint);
	    /* Do not retry: we have a negative answer */
	    *retry = 0;
	    break;
	case XACML_QUERY_PROTOCOL_ERROR :
	    log_fake_stderr(LOG_ERR, filedesc);
	    lcmaps_log(LOG_DEBUG,
		"%s: Failed to authorize and get a mapping"
		" with endpoint \"%s\" (on try #%u)\n",
		logstr, endpoint, try);
	    /* Retry */
	    *retry = 1;
	    break;
	default:
	    /* Invalid return code */
	    log_fake_stderr(LOG_ERR, filedesc);
	    lcmaps_log(LOG_ERR,
		"%s: Failed to authorize and get a mapping"
		" with endpoint (with odd error return from "
		"xacmlqueryscas()) \"%s\" (on try #%u)\n",
		logstr, endpoint, try);
	    /* Retry */
	    *retry = 1;
	    break;
    }

    /* Return the proper reduced error code */
    return returncode;
}

/******************************************************************************
Function:    parse_and_store_creddata
Description: Fetches, parses and stores the credential data returned in the
	     XACML response obligations. It checks the consistency and whether
	     it fulfills the minimal requirements. In case it parses correctly,
	     it stores the results in the LCMAPS framework.
Returns:     0 on success, -1 on error
******************************************************************************/
static int parse_and_store_creddata(const char *endpoint)   {
    uid_t *uid;
    gid_t *pgid;
    gid_t *sgid;
    size_t i,uid_cnt, pgid_cnt, sgid_cnt;

    /* Get resulting data */
    get_uid_list(&uid, &uid_cnt);
    get_pgid_list(&pgid, &pgid_cnt);
    get_sgid_list(&sgid, &sgid_cnt);

    /* Check consistency UID(s) */
    if (uid_cnt==0) {	/* No UIDs */
	lcmaps_log(LOG_ERR,
		"%s: XACML: No or invalid mapping result: no UIDs. "
		"This is an unknown state. Endpoint used: \"%s\"\n",
		__func__, endpoint);
	return -1;
    }
    if (uid_cnt>1)  { /* More than 1 uid: need to check consistency */
	for (i=1; i<uid_cnt; i++)	{
	    if (uid[i]!=uid[0]) {
		lcmaps_log(LOG_ERR,
			"%s: XACML: Inconsistent mapping result: "
			"multiple and different UIDs returned. "
			"This is an unknown state. Endpoint used: \"%s\"\n",
			__func__, endpoint);
		return -1;
	    }
	}
    }

    /* Check consistency primary GID(s) */
    if (pgid_cnt==0) {	/* No pGIDs */
	lcmaps_log(LOG_ERR,
		"%s: XACML: No or invalid mapping result: no pGIDs. "
		"This is an unknown state. Endpoint used: \"%s\"\n",
		__func__, endpoint);
	return -1;
    }
    if (pgid_cnt>1)  { /* More than 1 pGID: need to check consistency */
	/* Test for all cases except GUMS legacy behaviour of specifying both
	 * username+uidgid obligation: in that exceptional case, the UIDGID
	 * primary GID and the USERNAME primary GID are allowed to mismatch.
	 * In all other cases, they must match */
	if ( !(getOHUIDGIDFired() &&
	       getOHUsernameFired() &&
	       !getOHAccountFired()) )
	{
	    for (i=1; i<pgid_cnt; i++)	{
		if (pgid[i]!=pgid[0]) {
		    lcmaps_log(LOG_ERR,
			"%s: XACML: Inconsistent mapping result: "
			"multiple and different pGIDs returned. "
			"This is an unknown state. Endpoint used: \"%s\"\n",
			__func__, endpoint);
		    return -1;
		}
	    }
	}
    }

    /* Add (first) UID */
    if (addCredentialData(UID, uid))    {
	lcmaps_log(LOG_ERR, "%s: Error adding UID %d\n", __func__, uid[0]);
	return -1;
    }

    /* Add (first) pGID */
    if (addCredentialData(PRI_GID, pgid))    {
	lcmaps_log(LOG_ERR, "%s: Error adding pGID %d\n", __func__, pgid[0]);
	return -1;
    }

    /* secondary GID(s): no consistency checks, just add them all */
    for (i=0; i<sgid_cnt; i++)	{
	if (addCredentialData(SEC_GID, &(sgid[i]))) {
	    lcmaps_log(LOG_ERR,"%s: Error adding sGID %d\n", __func__, sgid[i]);
	    return -1;
	}
    }

    /* All done */
    return 0;
}



/******************************************************************************
Function:    lcmaps_scas_client_time_t_to_string
Description: converts an time_t into a string specified in the
             authorization interoperability workgroup document.
Parameters:
     time_t
     buf
     buf_len
Returns:
     int
******************************************************************************/
static int lcmaps_scas_client_time_t_to_string(time_t the_time,
					       char *buf,
					       size_t buf_len)
{
    struct tm res_tm;
    int rc;

    /* Time convert to UTC */
    if (!gmtime_r(&the_time, &res_tm)) {
        lcmaps_log(LOG_WARNING,
		"%s: Time conversion error from time_t to struct tm\n",
		__func__);
	buf[0]='\0';
	return -1;
    }

    rc=snprintf(buf, buf_len, "%04d-%02d-%02dT%02d:%02d:%02dZ",
		res_tm.tm_year + 1900,
		res_tm.tm_mon + 1,
		res_tm.tm_mday,
		res_tm.tm_hour,
		res_tm.tm_min,
		res_tm.tm_sec);

    if (rc<0 || (size_t)rc>=buf_len)	{
	lcmaps_log(LOG_WARNING,
		"%s: Time conversion error from struct tm to char *\n",
		__func__);
	buf[0]='\0';
	return -1;
    }
    
    return 0;
}


/**
 * Fills in all the subject attribute fields in subject_attr_t based on given
 * certificate chain. Memory to be freed using free_subject_attributes()
 */
static int fill_subject_attributes(subject_attr_t *d, STACK_OF(X509) *myChain)   {
    int i;
    X509 	*cert = NULL;
    int		depth=0, amount_of_CAs=0;
    time_t	notAfter=INT32_MAX, chain_notAfter=INT32_MAX;
    time_t	notBefore=INT32_MIN,chain_notBefore=INT32_MIN;
    CERTIFICATEPOLICIES *ca_policy_oids=NULL;
    POLICYINFO  *pol=NULL;

    if (myChain==NULL)	{
	lcmaps_log(LOG_DEBUG, "%s: myChain is empty, nothing to do\n",
		__func__);
	return 0;
    }

    if (d==NULL)   {
	lcmaps_log(LOG_ERR, "%s: internal error, called with NULL arguments\n",
		__func__);
	return -1;
    }

    /* If use_cert_chain_attribute is set, also put that in the struct */
    if (use_cert_chain_attribute)
	d->cert_chain=stack_of_x509_to_pem(myChain);

    /* count certificates in chain and determine notBefore/notAfter */
    depth = sk_X509_num(myChain);
    for (i = 0; i < depth; i++) {
	cert = sk_X509_value(myChain, i);

	/* Is it a CA ? */
	if (xacml_io_x509IsCA(cert))	{
	    lcmaps_log(LOG_DEBUG,"%s: CA cert at depth %d\n", __func__, i);
	    amount_of_CAs++;
	}

	/* Calculate the effective certificate chain validaty period */
	notAfter = xacml_io_asn1TimeToTimeT((char *)ASN1_STRING_data(X509_get_notAfter(cert)));
	if (notAfter < chain_notAfter)
	    chain_notAfter = notAfter;

	notBefore = xacml_io_asn1TimeToTimeT((char *)ASN1_STRING_data(X509_get_notBefore(cert)));
	if (notBefore > chain_notBefore)
	    chain_notBefore = notBefore;
    }
    if (chain_notBefore>=0)
	lcmaps_scas_client_time_t_to_string(chain_notBefore, d->validity_not_before, 21);
    if (chain_notAfter>=0)
	lcmaps_scas_client_time_t_to_string(chain_notAfter, d->validity_not_after, 21);

    /* Determine EEC */
    cert = sk_X509_value(myChain, depth - 1 - amount_of_CAs);
    d->subject_x509_id    =X509_NAME_oneline(X509_get_subject_name(cert),NULL,0);
    d->subject_x509_issuer=X509_NAME_oneline(X509_get_issuer_name (cert),NULL,0);
    d->certificate_serial_number = get_serialNumber(cert);

    /* Get Certificate Policies */
    ca_policy_oids = X509_get_ext_d2i(cert, NID_certificate_policies, 0, 0);
    if (ca_policy_oids) {
	d->numpols = sk_POLICYINFO_num(ca_policy_oids);
	if (d->numpols>0)   {
	    d->ca_policy_oid=(char**)malloc((size_t)(d->numpols)*sizeof(char*));
	    if (d->ca_policy_oid==NULL)	{
		lcmaps_log(LOG_ERR, "%s: out of memory\n", __func__);
		return -1;
	    }
	}
	for (i=0; i<d->numpols; i++)   {
	    d->ca_policy_oid[i]=(char*)calloc(sizeof(char),80);
	    if (d->ca_policy_oid[i]==NULL)	{
		/* tidy up */
		free(d->ca_policy_oid); d->ca_policy_oid=NULL; d->numpols=0;
		lcmaps_log(LOG_ERR, "%s: out of memory\n", __func__);
		return -1;
	    }
	    pol = sk_POLICYINFO_value(ca_policy_oids, i);
	    /* OBJ_obj2txt(3ssl) says that 80 should be more than
	     * enough */
	    OBJ_obj2txt(d->ca_policy_oid[i], 80, pol->policyid, 0);
	}
	CERTIFICATEPOLICIES_free(ca_policy_oids);
    }

    /* Determine CA cert */
    if (amount_of_CAs>0)    {
	cert = sk_X509_value(myChain, depth - amount_of_CAs);
	d->ca_issuer_dn = X509_NAME_oneline(X509_get_issuer_name(cert),NULL,0);
	/* serial should be positive according to RFC5280, page 4.1.2.2, not
	 * clear if that also applies to self-signed certificates. Dutchgrid
	 * currently has * a serial number 0 */
	d->ca_serial_number = get_serialNumber(cert);
    }

    return 0;
}


/**
 * Frees all the necessary memory in the subject_attr_t
 */
static void free_subject_attributes(subject_attr_t *d)  {
    int i;
    free(d->subject_x509_id); d->subject_x509_id=NULL;
    free(d->subject_x509_issuer); d->subject_x509_issuer=NULL;
    free(d->ca_issuer_dn); d->ca_issuer_dn=NULL;
    OPENSSL_free(d->certificate_serial_number);
    d->certificate_serial_number=NULL;
    OPENSSL_free(d->ca_serial_number);
    d->ca_serial_number=NULL;
    for (i=0; i<d->numpols; i++)
	free(d->ca_policy_oid[i]);
    d->numpols=0;
    free(d->ca_policy_oid);
    d->ca_policy_oid=NULL;
    free(d->cert_chain); d->cert_chain=NULL;
}

/**
 * Get serial number from given cert and return decimal integer representation
 * as character string. Returned value must be freed.
 */
static char *get_serialNumber(X509 *cert)   {
    ASN1_INTEGER *asn;
    BIGNUM *bn;
    char *serial;

    if (cert==NULL ||
	    (asn=X509_get_serialNumber(cert))==NULL ||  /* -> ASN1_INTEGER */
	    (bn=ASN1_INTEGER_to_BN(asn, NULL))==NULL)   /* -> BIGNUM */
	return NULL;

    /* Get char* representation */
    serial=BN_bn2dec(bn);
    BN_free(bn);

    return serial;
}

/**
 * Convert stack of X509 into PEM. Returned value must be freed.
 */
static char *stack_of_x509_to_pem(STACK_OF(X509) *myChain) {
    BUF_MEM *bptr;
    size_t length;
    int i,blocksize,depth,rc;
    BIO *membio;
    X509 *cert;
    char *pem=NULL;
   
    /* Create memory bio */
    if ((membio=BIO_new(BIO_s_mem()))==NULL)	{
	lcmaps_log(LOG_ERR,"%s: cannot create memory BIO\n", __func__);
	return NULL;
    }

    /* Write certificates to bio */
    depth = sk_X509_num(myChain);
    for (i = 0; i < depth; i++) {
	cert = sk_X509_value(myChain, i);
	if (!PEM_write_bio_X509(membio, cert))	{
	    lcmaps_log(LOG_ERR,"%s: cannot write to BIO (cert %d)\n",
		    __func__, i);
	    goto cleanup;
	}
    }
    /* Get amount of data */
    BIO_get_mem_ptr(membio, &bptr);
    length = bptr->length;

    /* Allocate array: need one extra for the trailing NULL character */
    if ((pem=(char*)calloc(length+1,1))==NULL)  {
	lcmaps_log(LOG_ERR, "%s: out of memory\n",__func__);
	goto cleanup;
    }

    /* We can do better, but pemstring longer than INT_MAX are in any case
     * worrisome */
    if (length>INT_MAX)	{
	lcmaps_log(LOG_ERR,
		"%s: length of pemstring is too large to dump\n",
		__func__);
	goto cleanup;
    }

    blocksize=(int)length;
    /* Dump pemstring into array */
    if ((rc=BIO_read(membio,pem,blocksize))!=blocksize)   {
	lcmaps_log(LOG_ERR,
		"%s: error reading PEM from BIO, "
		"BIO_read() returned %d instead of %d\n",
		__func__, rc, blocksize);
	free(pem); pem=NULL;
	goto cleanup;
    }

cleanup:
    BIO_free(membio);
    return pem;
}
